from pathlib import Path
from jinja2 import Environment, PackageLoader, StrictUndefined

from pystencils.backends.cbackend import generate_c, get_headers
from pystencils.astnodes import KernelFunction, PragmaBlock
from pystencils.enums import Backend
from pystencils.data_types import get_base_type
from pystencils.sympyextensions import prod
from pystencils.integer_functions import modulo_ceil

import numpy as np


def generate_benchmark(kernel_ast: KernelFunction,
                       path: Path = None,
                       *,
                       dialect: Backend = Backend.C) -> None:
    if path is None:
        path = Path('.')
    else:
        path.mkdir(parents=True, exist_ok=True)
    src_path = path / 'src'
    src_path.mkdir(parents=True, exist_ok=True)
    include_path = path / 'include'
    include_path.mkdir(parents=True, exist_ok=True)

    kernel_name = kernel_ast.function_name

    header = kernel_header(kernel_ast, dialect)
    with open(include_path / f'{kernel_name}.h', 'w+') as f:
        f.write(header)

    source = kernel_source(kernel_ast, dialect)
    with open(src_path / f'{kernel_name}.c', 'w+') as f:
        f.write(source)

    with open(src_path / 'main.c', 'w+') as f:
        f.write(kernel_main(kernel_ast))


def kernel_main(kernel: KernelFunction, timing: bool = False):
    """
    Return C code of a benchmark program for the given kernel.

    Args:
        kernel: the pystencils AST object as returned by create_kernel
        timing: add timing output to the code, prints time per iteration to stdout
    Returns:
        C code as string
    """
    kernel_name = kernel.function_name
    accessed_fields = {f.name: f for f in kernel.fields_accessed}
    constants = []
    fields = []
    call_parameters = []
    for p in kernel.get_parameters():
        if not p.is_field_parameter:
            constants.append((p.symbol.name, str(p.symbol.dtype)))
            call_parameters.append(p.symbol.name)
        else:
            assert p.is_field_pointer, "Benchmark implemented only for kernels with fixed loop size"
            field = accessed_fields[p.field_name]
            dtype = str(get_base_type(p.symbol.dtype))
            np_dtype = get_base_type(p.symbol.dtype).numpy_dtype
            size_data_type = np_dtype.itemsize

            dim0_size = field.shape[-1]
            dim1_size = np.prod(field.shape[:-1])
            elements = prod(field.shape)

            if kernel.instruction_set:
                align = kernel.instruction_set['width'] * size_data_type
                padding_elements = modulo_ceil(dim0_size, kernel.instruction_set['width']) - dim0_size
                padding_bytes = padding_elements * size_data_type
                ghost_layers = max(max(kernel.ghost_layers))

                size = dim1_size * padding_bytes + np.prod(field.shape) * size_data_type

                assert align % np_dtype.itemsize == 0
                offset = ((dim0_size + padding_elements + ghost_layers) % kernel.instruction_set['width']) * size_data_type

                fields.append((p.field_name, dtype, elements, size, offset, align))
                call_parameters.append(p.field_name)
            else:
                size = elements * size_data_type
                fields.append((p.field_name, dtype, elements, size, 0, 0))
                call_parameters.append(p.field_name)

    includes = f'#include "{kernel_name}"\n'
    jinja_context = {
        'kernel_code': generate_c(kernel, dialect=Backend.C),
        'kernel_name': kernel_name,
        'fields': fields,
        'constants': constants,
        'call_argument_list': ",".join(call_parameters),
        'includes': includes,
        'timing': timing,
    }

    env = Environment(loader=PackageLoader('pystencils_benchmark'), undefined=StrictUndefined)
    main = env.get_template('main.c').render(**jinja_context)
    return main


def kernel_header(kernel_ast: KernelFunction, dialect: Backend = Backend.C) -> str:
    function_signature = generate_c(kernel_ast, dialect=dialect, signature_only=True)

    jinja_context = {
        'function_signature': function_signature,
    }

    env = Environment(loader=PackageLoader('pystencils_benchmark'), undefined=StrictUndefined)
    header = env.get_template('kernel.h').render(**jinja_context)
    return header


def kernel_source(kernel_ast: KernelFunction, dialect: Backend = Backend.C) -> str:
    kernel_name = kernel_ast.function_name
    function_source = generate_c(kernel_ast, dialect=dialect)
    headers = get_headers(kernel_ast)
    headers.add(f'"{kernel_name}.h"')

    jinja_context = {
        'function_source': function_source,
        'headers': sorted(headers),
    }

    env = Environment(loader=PackageLoader('pystencils_benchmark'), undefined=StrictUndefined)
    source = env.get_template('kernel.c').render(**jinja_context)
    return source
